
def aead(impl):
    from drysponge.base import DrySponge
    import sys
    import binascii

    if (len(sys.argv) > 8) | (len(sys.argv) < 6):
        print("ERROR: need at least 5 arguments: operation,key,nonce,message,associatedData,[staticData],[verbosity]")
        print("""
operation:
    e,enc,encrypt
    d,dec,decrypt

verbosity: from 0 to 5

examples:
    no message, no associated data: e 000102030405060708090A0B0C0D0E0F 000102030405060708090A0B0C0D0E0F "" ""
    message, no associated data:    e 000102030405060708090A0B0C0D0E0F 000102030405060708090A0B0C0D0E0F "AABB" ""
    no message, associated data:    e 000102030405060708090A0B0C0D0E0F 000102030405060708090A0B0C0D0E0F "" "AABB"
    message and associated data:    e 000102030405060708090A0B0C0D0E0F 000102030405060708090A0B0C0D0E0F "AABB" "CCDD"
    same with verbose output:       e 000102030405060708090A0B0C0D0E0F 000102030405060708090A0B0C0D0E0F "AABB" "CCDD" 2
        """)
        exit()
    op = sys.argv[1]
    key = binascii.unhexlify(sys.argv[2])
    nonce = binascii.unhexlify(sys.argv[3])
    message = binascii.unhexlify(sys.argv[4])
    associatedData = binascii.unhexlify(sys.argv[5])
    staticData = None
    verbosityIdx = None
    if len(sys.argv) > 6 :
        if len(sys.argv[6]) > 1:
            staticData = binascii.unhexlify(sys.argv[6])
        else:
            assert(len(sys.argv) == 7)
            verbosityIdx = 6
    if len(sys.argv) == 8 :
        assert(verbosityIdx is None)
        verbosityIdx = 7
    if verbosityIdx is not None:
        impl.Verbose(int(sys.argv[verbosityIdx]))
    if op in ['e','enc','encrypt']:
        out = impl.encrypt(key,nonce,message,associatedData,staticData)
    elif op in ['d','dec','decrypt']:
        out = impl.decrypt(key,nonce,message,associatedData,staticData)
    else:
        out = impl.encrypt(key,nonce,message,associatedData,staticData)
        impl.decrypt(key,nonce,out,associatedData,staticData)
    DrySponge.print_bytes_as_hex(out)



# DrySponge: Sponge that does not leak
# Sebastien Riou, January 5th 2019
# Convention:
#  width means length in bits
#  size means length in bytes
#
# version=2: changes in key setup for the minium key size of each security level
import sys
import binascii
import os
import copy
import random

class DrySponge(object):

    SPY_NONE = 0
    SPY_ALG_IO = 1
    SPY_F_IO = 2
    SPY_ROUND_IO = 3
    SPY_FULL = 4

    DS_S = 2
    DS_D = 1
    DS_A = 2
    DS_M = 3

    PRINT_C99 = os.getenv('PRINT_C99',0)

    def __init__(self,params):
        assert(0==(params.MinKeyWidth%8))
        assert(0==(params.RateWidth%8))
        assert(0==(params.CapacityWidth%8))
        self.params = params
        self.__Version = params.version
        assert(self.__Version in [1,2])
        self.__VerboseLevel = 0
        self.__MinKeyWidth = params.MinKeyWidth
        self.__NonceWidth = params.NonceWidth
        self.__RateWidth = params.RateWidth
        self.__XWidth = params.xwords*32
        self.__CapacityWidth = params.CapacityWidth
        self.__MixPhase = params.MixPhase
        self.__InitRounds = params.InitRounds
        self.__Rounds = params.Rounds
        #self.__FinalRounds = params.FinalRounds
        self.__CoreRound = params.CoreRound
        self.__DomainSeparator = params.DomainSeparator
        self.__AccumulateFactor = params.AccumulateFactor
        self.__BLOCKSIZE = params.RateWidth // 8
        self.__CL_SBOUND = 0
        self.__CL_EBOUND = self.__BLOCKSIZE
        self.__CH_SBOUND = self.__CL_EBOUND
        self.__CH_EBOUND = self.__CH_SBOUND + self.__BLOCKSIZE
        self.__C_EBOUND = params.CapacityWidth // 8
        self.__X_SBOUND  = self.__C_EBOUND
        self.__X_EBOUND  = self.__X_SBOUND + params.xwords*4

        self.__R_SBOUND  = self.__X_EBOUND
        self.__R_EBOUND  = self.__R_SBOUND + self.__BLOCKSIZE

        self.__StateWidth = params.RateWidth + params.CapacityWidth + params.xwords*32
        self.__CST_H = DrySponge.int_to_bytes(0x243F6A8885A308D313198A2E03707344A4093822299F31D0082EFA98EC4E6C89452821E638D01377BE5466CF34E90C6CC0AC29B7C97C50DD3F84D5B5B54709179216D5D98979FB1BD1310BA698DFB5AC2FFD72DBD01ADFB7B8E1AFED6A267E96BA7C9045F12C7F9924A19947B3916CF70801F2E2858EFC16636920D871574E69,1024)
        self.__CST_H.reverse()
        #for i in range(0,len(self.__CST_H)//8):
        #    w=DrySponge.bytes_to_int(self.__CST_H[8*i:8*(i+1)])
        #    print("0x%016x,"%w)
        #for i in range(0,len(self.__CST_H)//8):
        #    for j in range(0,8):
        #        print("0x%02x,"%self.__CST_H[i*8+j],end="")
        #    print()
        #things that shall be initialized by __init()
        self.__State = None
        self.__fcnt = None

    def __init(self):
        self.__State = DrySponge.int_to_bytes(0,self.__StateWidth)
        self.__fcnt = 0

    def __c(self):
        return copy.deepcopy(self.__State[self.__CL_SBOUND:self.__C_EBOUND])

    def __clh(self):
        return copy.deepcopy(self.__State[self.__CL_SBOUND:self.__CH_EBOUND])

    def __cl(self):
        return copy.deepcopy(self.__State[self.__CL_SBOUND:self.__CL_EBOUND])

    def __ch(self):
        return copy.deepcopy(self.__State[self.__CH_SBOUND:self.__CH_EBOUND])

    def __x(self):
        return copy.deepcopy(self.__State[self.__X_SBOUND:self.__X_EBOUND])

    def __r(self):
        return copy.deepcopy(self.__State[self.__R_SBOUND:self.__R_EBOUND])

    @staticmethod
    def limit_len(data,length):
        if len(data) > length:
            data = copy.deepcopy(data[0:length])
        return data

    def __set_c(self,val):
        self.__State[self.__CL_SBOUND:self.__C_EBOUND] = self.limit_len(val,self.__CapacityWidth//8)

    def __set_x(self,val):
        self.__State[self.__X_SBOUND:self.__X_EBOUND] = self.limit_len(val,self.__XWidth//8)

    def __set_r(self,val):
        self.__State[self.__R_SBOUND:self.__R_EBOUND] = self.limit_len(val,self.__RateWidth//8)

    def Verbose(self,level=None):
        if level is not None:
            self.__VerboseLevel = level
            self.params.spy = (self.SPY_FULL<=level)
            self.params.spy_all = (self.SPY_FULL<level)
        return self.__VerboseLevel

    @staticmethod
    def bytes_as_hexstr(block, end=""):
        out=""
        for b in block:
            out += "%02X"%b
        out += end
        return out

    @staticmethod
    def bytes_as_hexc99(block, end=""):
        out=""
        sep=""
        for b in block:
            out += sep+"0x%02X"%b
            sep=","
        out += end
        return out

    @staticmethod
    def bytes_as_hex(block, end=""):
        if DrySponge.PRINT_C99:
            return DrySponge.bytes_as_hexc99(block,end)
        else:
            return DrySponge.bytes_as_hexstr(block,end)

    @staticmethod
    def print_bytes_as_hex(block, end="\n"):
        print(DrySponge.bytes_as_hex(block),end=end)

    def __print_state(self, msg=None, level=0):
        if self.__VerboseLevel >= level:
            if msg is not None:
                print(msg)
            linesize = max(32,self.__BLOCKSIZE)
            for i in range(0,(self.__CapacityWidth//8+linesize-1) // linesize):
                self.__print("   C[%2d] ="%i,self.__c()[i*linesize:(i+1)*linesize],level)
            for i in range(0,(self.__XWidth//8+linesize-1) // linesize):
                self.__print("   X[%2d] ="%i,self.__x()[i*linesize:(i+1)*linesize],level)
            self.__print("       R =",self.__r(),level)

    def __print(self, msg=None, data=None, level=0):
        if self.__VerboseLevel >= level:
            data_str = ""
            if data is not None:
                data_str = self.bytes_as_hex(data)
            if msg is None:
                msg = data_str
            else:
                msg = msg + " " + data_str
            print(msg)

    @staticmethod
    def int_to_bytes(a,width=8):
        return bytearray(a.to_bytes((width+7)//8, byteorder='little'))

    @staticmethod
    def bytes_to_int(a):
        if 0 == len(a):
            return 0
        return int.from_bytes(a, byteorder='little', signed=False)

    def __rotr32(self,val, r):
        return ((val >> r) ^ (val << (32-r))) % (1 << 32)

    def __xor(self,a,b):
        return bytearray(bytes([ (x ^ y) for (x,y) in zip(a, b) ]))

    @staticmethod
    def swap32(a,b,s):
        for i in range(0,4):
            tmp=s[a+i]
            s[a+i]=s[b+i]
            s[b+i]=tmp

    def __set_key_v1(self,key):
        minkeysize = self.__MinKeyWidth // 8;
        xsize = self.__XWidth//8
        xwords = self.__XWidth // 32
        csize = self.__CapacityWidth//8
        midkeysize = minkeysize+xsize;
        fullkeysize = csize+xsize
        assert(len(key)>=minkeysize)
        assert(len(key)<=fullkeysize)
        k = copy.deepcopy(key)
        if fullkeysize==len(key):
            self.__set_c(k[:csize])
            self.__set_x(k[csize:])
        else:
            kc = copy.deepcopy(k[:minkeysize])
            while len(kc) < csize:
                kc += copy.deepcopy(kc)
            self.__set_c(kc[:csize])
            if midkeysize==len(key):
                self.__set_x(k[minkeysize:])
            else:
                assert(minkeysize==len(key))
                self.__print_state("   Key setup entry:",self.SPY_ROUND_IO)
                self.__set_c(self.__CoreRound(self.__c(),0))
                modified=1
                cnt=1
                while modified:
                    modified=0
                    c = self.__c()
                    for i in range(0,xwords-1):
                        for j in range(i+1,xwords):
                            if c[i*4:i*4+4] == c[j*4:j*4+4]:
                                self.__print_state("   X fixup entry %d:"%cnt,self.SPY_ROUND_IO)
                                self.__set_c(self.__CoreRound(self.__c(),0))
                                modified=1
                                cnt+=1
                                break
                        if modified:
                            break
                self.__set_x(c[:xsize])
                assert(len(key)>=xsize)
                c[:xsize] = copy.deepcopy(key[:xsize])
                self.__set_c(c)
        x = self.__x()
        for i in range(0,xwords-1):
            for j in range(i+1,xwords):
                assert(x[i*4:i*4+4] != x[j*4:j*4+4])

    def __set_key_v2(self,key):
        minkeysize = self.__MinKeyWidth // 8;
        xsize = self.__XWidth//8
        xwords = self.__XWidth // 32
        csize = self.__CapacityWidth//8
        midkeysize = minkeysize+xsize;
        fullkeysize = csize+xsize
        assert(len(key)>=minkeysize)
        assert(len(key)<=fullkeysize)
        k = copy.deepcopy(key)
        if fullkeysize==len(key):
            c=k[:csize]
            x=k[csize:]
        else:
            c = copy.deepcopy(k[:minkeysize])
            while len(c) < csize:
                c += copy.deepcopy(c)
            c=c[:csize]
            if midkeysize==len(key):
                x=k[minkeysize:]
            else:
                assert(minkeysize==len(key))
                assert(len(key)>=xsize)
                self.__set_c(c)
                self.__print_state("   Key setup entry:",self.SPY_ROUND_IO)
                x=self.__CoreRound(self.__c(),0)[:xsize]
        assert(len(c)==csize)
        assert(len(x)==xsize)
        self.__set_c(c)
        x = bytearray(x)
        match=0
        for i in range(0,xwords-1):
            for j in range(i+1,xwords):
                if x[i*4:i*4+4] == x[j*4:j*4+4]:
                    match=1
        mask = 0xFFFFFFFF << (match * xwords.bit_length())
        for i in range(0,xwords):
            xi = self.bytes_to_int(x[i*4:i*4+4])
            xi = (xi & mask) | (i & ~mask)
            x[i*4:i*4+4] = self.int_to_bytes(xi,32)
        self.__set_x(x)
        x = self.__x()
        for i in range(0,xwords-1):
            for j in range(i+1,xwords):
                assert(x[i*4:i*4+4] != x[j*4:j*4+4])

    def __set_key(self,key):
        if self.__Version == 1:
            self.__set_key_v1(key)
        else:
            self.__set_key_v2(key)

    def __g(self,caller_is_F=0):
        if 0==caller_is_F:
            self.__print_state("   F/G entry %d (G):"%self.__fcnt,self.SPY_F_IO)
        self.__fcnt += 1
        self.__set_r(DrySponge.int_to_bytes(0,self.__BLOCKSIZE*8))
        blocksize32 = (self.__BLOCKSIZE+3)//4 #rounded up
        for j in range(0,self.params.rounds):
            self.__print_state("   CoreRound entry %d:"%j,self.SPY_ROUND_IO)
            self.__set_c(self.__CoreRound(self.__c(),j))
            r = self.__r()
            self.__print_state("   Accumulate entry %d:"%j,self.SPY_FULL)
            for k in range(0,self.__AccumulateFactor):
                cpart = self.__c()[k*4*blocksize32:]
                cpart = cpart[0:self.__BLOCKSIZE]
                cpart2 = cpart[k*4:] + cpart[0:k*4]
                self.__print("accumulate in[%d] ="%k,cpart2,self.SPY_FULL)
                r = self.__xor(r,cpart2)
            self.__set_r(r)
        return self.__r()

    def __f(self,i,padded=None,domain=None,finalize=None):
        assert(len(i)==self.__BLOCKSIZE)
        if padded is not None:
            self.__DomainSeparator(padded,domain,finalize)
            self.__print_state("   F/G entry %d (F with DS): padded=%d, domain=%d, finalize=%d"%(self.__fcnt,padded,domain,finalize),self.SPY_F_IO)
        else:
            self.__print_state("   F/G entry %d (F):"%(self.__fcnt),self.SPY_F_IO)

        self.__print("       I =",i,self.SPY_F_IO)
        (c,x)=self.__MixPhase(self.__c(),self.__x(),i)
        self.__set_c(c)
        self.__set_x(x)
        self.__print_state("   After mix phase:",self.SPY_ROUND_IO)
        return self.__g(1)

    def __padding(self,lastblock):
        padded = 0
        pad = binascii.unhexlify(b'01')
        for i in range(len(lastblock),self.__BLOCKSIZE):
            lastblock += pad
            pad = binascii.unhexlify(b'00')
            padded = 1
        return padded

    def key_size(self):
        return self.__MinKeyWidth // 8

    def fastkey_size(self):
        return (self.__MinKeyWidth+self.__XWidth) // 8

    def fullkey_size(self):
        return (self.__CapacityWidth+self.__XWidth) // 8

    def block_size(self):
        return self.__BLOCKSIZE

    def nonce_size(self):
        return self.__NonceWidth // 8

    def tag_size(self):
        return self.__MinKeyWidth // 8

    def __swap32(self,a,b):
        c=self.__c()
        for i in range(0,4):
            tmp = c[a*4+i]
            c[a*4+i] = c[b*4+i]
            c[b*4+i] = tmp
        self.__set_c(c)

    def __absorb_only(self,ad,domain,final,paddedmsg):
        alen = len(ad)
        a = (alen + self.__BLOCKSIZE -1) // self.__BLOCKSIZE

        if a>0:
            lastblock = bytearray(copy.deepcopy(ad[(a-1)*self.__BLOCKSIZE:]))
        else:
            a=1
            lastblock = bytearray(b'')

        apad = self.__padding(lastblock)
        if apad:
            self.__print(paddedmsg,ad[:(a-1)*self.__BLOCKSIZE]+lastblock,self.SPY_ALG_IO)

        for i in range(0,a-1):
            ai = ad[i*self.__BLOCKSIZE:(i+1)*self.__BLOCKSIZE]
            self.__f(ai)

        self.__f(lastblock,apad,domain,final)

    def hash(self,message):
        self.__print("Hashing %d bytes message:"%len(message),message,self.SPY_ALG_IO)
        self.__init()
        self.params.rounds = self.params.Rounds
        self.__set_key(self.__CST_H[:(self.__MinKeyWidth+self.__XWidth)//8])
        self.__absorb_only(message,self.DS_S,1,"   Padded Message: ")
        out = self.__r()
        for i in range(1,2*((self.__MinKeyWidth + self.__RateWidth -1 ) // self.__RateWidth) ):
            out += self.__g()
        out = out[:2*self.__MinKeyWidth // 8]
        self.__print_state("   Final state:",self.SPY_F_IO)
        self.__print("   Digest:",out,self.SPY_ALG_IO)
        return out

    def encrypt(self,key,nonce,message=None,associated_data=None,static_data=None):
        self.params.rounds = self.params.Rounds
        if message is None:
            message = bytearray(binascii.unhexlify(b''))
        if associated_data is None:
            associated_data = bytearray(binascii.unhexlify(b''))
        if static_data is None:
            static_data = bytearray(binascii.unhexlify(b''))
        self.__print(
            "Encrypting %d bytes message with %d bytes of associated data and %d bytes of static data"%(len(message),len(associated_data),len(static_data)),
            None,self.SPY_ALG_IO)
        self.__print(
            "Key size: %d bytes"%(len(key)),
            None,self.SPY_ALG_IO)
        self.__print("   Key:            ",key,self.SPY_ALG_IO)
        self.__print("   Nonce:          ",nonce,self.SPY_ALG_IO)
        if len(static_data):
            self.__print("   Static data:    ",static_data,self.SPY_ALG_IO)
        if len(associated_data):
            self.__print("   Associated data:",associated_data,self.SPY_ALG_IO)
        if len(message):
            self.__print("   Message:        ",message,self.SPY_ALG_IO)
        self.__init()
        self.__set_key(key)
        #assert(0==(len(nonce)%self.block_size()))
        assert(12<=len(nonce))
        final=0
        if (0==len(message)) & (0==len(associated_data)):
            final=1

        #static data
        if len(static_data):
            self.__absorb_only(static_data,self.DS_S,final,"   Padded S. data: ")
            final=0
            sdk = self.__r()
            for i in range(1,(self.__CapacityWidth + self.__RateWidth -1 )// self.__RateWidth):
                sdk += self.__g()
            self.__set_c(sdk)

        self.params.rounds = self.params.InitRounds
        #diversification
        if self.__NonceWidth > self.__RateWidth:
            self.__f(nonce[0:self.__BLOCKSIZE],0,self.DS_D,final)
            for i in range(1,self.__NonceWidth//self.__RateWidth):
                self.__f(nonce[i*self.__BLOCKSIZE:(i+1)*self.__BLOCKSIZE])
        else:
            block = bytearray(binascii.unhexlify(b'00')*self.block_size())
            block[0:self.nonce_size()] = nonce
            self.__f(block,0,self.DS_D,final)

        self.params.rounds = self.params.Rounds

        #associated data
        if len(associated_data):
            final = 0==len(message)
            self.__absorb_only(associated_data,self.DS_A,final,"   Padded A. data: ")

        mlen = len(message)
        m = (mlen + self.__BLOCKSIZE -1) // self.__BLOCKSIZE
        mpad = 0

        out = bytearray(binascii.unhexlify(b''))

        if m>0:
            mlastblock = bytearray(copy.deepcopy(message[(m-1)*self.__BLOCKSIZE:]))
            mpad = self.__padding(mlastblock)
            if mpad:
                self.__print("   Padded Message: ",message[:(m-1)*self.__BLOCKSIZE]+mlastblock,self.SPY_ALG_IO)

            for i in range(0,m-1):
                pi = message[i*self.__BLOCKSIZE:(i+1)*self.__BLOCKSIZE]
                out += self.__xor(self.__r(),pi)
                self.__f(pi)

            out += self.__xor(self.__r(),mlastblock)
            out = out[:len(message)]
            self.__f(mlastblock,mpad,self.DS_M,1)

        tag = self.__r()
        for i in range(1,(self.__MinKeyWidth + self.__RateWidth -1 )// self.__RateWidth):
            tag += self.__g()
        tag = tag[:self.__MinKeyWidth // 8]
        self.__print_state("   Final state:",self.SPY_F_IO)
        self.__print("   CipherText:     ",out,self.SPY_ALG_IO)
        self.__print("   Tag:            ",tag,self.SPY_ALG_IO)
        #self.__print("   CipherText|Tag: ",out+tag,self.SPY_ALG_IO)
        return out+tag

    def decrypt(self,key,nonce,ciphertext,associated_data=None,static_data=None):
        self.params.rounds = self.params.Rounds
        self.__init()
        if associated_data is None:
            associated_data = bytearray(binascii.unhexlify(b''))
        if static_data is None:
            static_data = bytearray(binascii.unhexlify(b''))
        self.__print(
            "Decrypting %d bytes ciphertext with %d bytes of associated data and %d bytes of static data"%(len(ciphertext),len(associated_data),len(static_data)),
            None,self.SPY_ALG_IO)
        self.__print(
            "Key size: %d bytes"%(len(key)),
            None,self.SPY_ALG_IO)
        message = ciphertext[:-self.tag_size()]
        expected_tag = ciphertext[-self.tag_size():]
        self.__print("   Key:            ",key,self.SPY_ALG_IO)
        self.__print("   Nonce:          ",nonce,self.SPY_ALG_IO)
        if len(static_data):
            self.__print("   Static data:    ",static_data,self.SPY_ALG_IO)
        if len(associated_data):
            self.__print("   Associated data:",associated_data,self.SPY_ALG_IO)
        if len(message):
            self.__print("   CipherText Msg: ",message,self.SPY_ALG_IO)
        self.__print("   CipherText Tag: ",expected_tag,self.SPY_ALG_IO)
        assert(len(key)>=self.key_size())
        self.__set_key(key)
        #assert(0==(len(nonce)%self.block_size()))
        assert(96>=len(nonce))
        final=0
        if (0==len(message)) & (0==len(associated_data)):
            final=1

        #static data
        if len(static_data):
            self.__absorb_only(static_data,self.DS_S,final,"   Padded S. data: ")
            final=0
            sdk = self.__r()
            for i in range(1,(self.__CapacityWidth + self.__RateWidth -1 )// self.__RateWidth):
                sdk += self.__g()
            self.__set_c(sdk)

        self.params.rounds = self.params.InitRounds
        #diversification
        if self.__NonceWidth > self.__RateWidth:
            nloops = (self.__NonceWidth + self.__RateWidth -1)//self.__RateWidth
            for i in range(0,nloops-1):
                self.__f(nonce[i*self.__BLOCKSIZE:(i+1)*self.__BLOCKSIZE])
            self.__f(nonce[(nloops-1)*self.__BLOCKSIZE:nloops*self.__BLOCKSIZE],0,self.DS_D,final)
        else:
            block = bytearray(binascii.unhexlify(b'00')*self.block_size())
            block[0:self.nonce_size()] = nonce
            self.__f(block,0,self.DS_D,final)
        self.params.rounds = self.params.Rounds

        #associated data
        if len(associated_data):
            final = 0==len(message)
            self.__absorb_only(associated_data,self.DS_A,final,"   Padded A. data: ")

        mlen = len(message)
        m = (mlen + self.__BLOCKSIZE -1) // self.__BLOCKSIZE
        mpad = 0
        out = bytearray(binascii.unhexlify(b''))

        if m>0:
            mlastblock = bytearray(copy.deepcopy(message[(m-1)*self.__BLOCKSIZE:]))
            for i in range(0,m-1):
                ci = message[i*self.__BLOCKSIZE:(i+1)*self.__BLOCKSIZE]
                pi = self.__xor(self.__r(),ci)
                out += pi
                self.__f(pi)

            mlastblock = self.__xor(self.__r(),mlastblock)
            mpad = self.__padding(mlastblock)
            out += mlastblock
            if mpad:
                self.__print("   Padded Message: ",out,self.SPY_ALG_IO)
            out = out[:len(message)]
            self.__DomainSeparator(mpad,self.DS_M,1)
            self.__f(mlastblock)

        tag = self.__r()
        for i in range(1,(self.__MinKeyWidth + self.__RateWidth -1 )//self.__RateWidth):
            tag += self.__g()
        tag = tag[:self.__MinKeyWidth // 8]
        self.__print_state("   Final state:",self.SPY_F_IO)
        if tag != expected_tag:
            raise ValueError('authentication tags different')
        self.__print("   Message:        ",out,self.SPY_ALG_IO)
        return out



import binascii
import copy

class Gascon(object):
    def __init__(self,nw,rounds):
        assert(1==(nw%2)) #sbox not bijective if nw is even
        assert(nw in [5,9])
        self.__nw = nw
        self.__rounds = rounds
        assert(self.__rounds <= 12)

    @staticmethod
    def sbox(i,nw):
        S = [0]*nw
        for b in range(0,nw):
            S[b] = (i >> b) & 1
        Gascon.sboxes(S,nw)
        r = 0
        for b in range(0,nw):
            r |= (S[b] & 1)<<b
        return r

    @staticmethod
    def __dbg(S,spy=True):
        if spy:
            for i in range(0,len(S)):
                print("%d: %016x"%(i,S[i]))
            print()

    @staticmethod
    def sboxes(S,nw):
        #Gascon.__dbg(S)
        mid = nw//2
        for i in range(0,mid+1):
            dst = 2*i
            src = (nw+dst-1) % nw
            S[dst] ^= S[src]
        #Gascon.__dbg(S)
        T = [(S[i] ^ 0xFFFFFFFFFFFFFFFF) & S[(i+1)%nw] for i in range(nw)]
        #Gascon.__dbg(T)
        for i in range(nw):
            S[i] ^= T[(i+1)%nw]
        #Gascon.__dbg(S)
        for i in range(0,mid+1):
            src = (2*i) % nw
            dst = (src+1) % nw
            S[dst] ^= S[src]
        S[mid] ^= 0XFFFFFFFFFFFFFFFF
        #Gascon.__dbg(S)

    def round(self,S,round,spy=False):
        """
        Gascon core permutation.
        S: Gascon state, a list of nw 64-bit integers
        round: round to perform
        returns nothing, updates S
        """
        assert(round <= 12)
        r = 12-self.__rounds+round
        mid = self.__nw//2
        # --- add round constants ---
        #round_constants = [
        #    0x0000000c0000000c,
        #    0x0000000c00000009,
        #    0x000000090000000c,
        #    0x0000000900000009,
        #    0x0000000c00000006,
        #    0x0000000c00000003,
        #    0x0000000900000006,
        #    0x0000000900000003,
        #    0x000000060000000c,
        #    0x0000000600000009,
        #    0x000000030000000c,
        #    0x0000000300000009
        #    ]
        #S[mid] ^= round_constants[r]
        S[mid] ^= ((0xF-r)<<4) | r
        if spy:
            print("rounds=%u, round=%u, r=%u"%(self.__rounds,round,r))
        self.__dbg(S,spy)
        # --- substitution layer ---
        self.sboxes(S,self.__nw)
        self.__dbg(S,spy)
        # --- linear diffusion layer ---
        S[0] ^= self.rotr64_interleaved(S[0], 19) ^ self.rotr64_interleaved(S[0], 28)
        S[1] ^= self.rotr64_interleaved(S[1], 61) ^ self.rotr64_interleaved(S[1], 38)
        S[2] ^= self.rotr64_interleaved(S[2],  1) ^ self.rotr64_interleaved(S[2],  6)
        S[3] ^= self.rotr64_interleaved(S[3], 10) ^ self.rotr64_interleaved(S[3], 17)
        S[4] ^= self.rotr64_interleaved(S[4],  7) ^ self.rotr64_interleaved(S[4], 40)
        if self.__nw > 5:
            S[5] ^= self.rotr64_interleaved(S[5], 31) ^ self.rotr64_interleaved(S[5], 26)
            S[6] ^= self.rotr64_interleaved(S[6], 53) ^ self.rotr64_interleaved(S[6], 58)
            S[7] ^= self.rotr64_interleaved(S[7],  9) ^ self.rotr64_interleaved(S[7], 46)
            S[8] ^= self.rotr64_interleaved(S[8], 43) ^ self.rotr64_interleaved(S[8], 50)
        #if self.__nw > 9:
        #    S[ 9] ^= self.rotr64_interleaved(S[ 9], ?) ^ self.rotr64_interleaved(S[ 9], ?)
        #    S[10] ^= self.rotr64_interleaved(S[10], ?) ^ self.rotr64_interleaved(S[10], ?)
        #    S[11] ^= self.rotr64_interleaved(S[11], ?) ^ self.rotr64_interleaved(S[11], ?)
        #    S[12] ^= self.rotr64_interleaved(S[12], ?) ^ self.rotr64_interleaved(S[12], ?)
        #    S[13] ^= self.rotr64_interleaved(S[13], ?) ^ self.rotr64_interleaved(S[13], ?)
        #    S[14] ^= self.rotr64_interleaved(S[14], ?) ^ self.rotr64_interleaved(S[14], ?)
        self.__dbg(S,spy)
        #if spy:
        #    exit()

    @staticmethod
    def rotr64(val, r):
        return ((val >> r) ^ (val << (64-r))) % (1 << 64)

    @staticmethod
    def rotr32(val, r):
        return ((val >> r) ^ (val << (32-r))) % (1 << 32)

    @staticmethod
    def rotr64_interleaved(val, r):
        i = [val & 0xFFFFFFFF, val>>32]
        shift2 = r//2
        if r & 1:
            tmp = Gascon.rotr32(i[1],shift2);
            i[1] = Gascon.rotr32(i[0],(shift2+1)%32);
            i[0] = tmp;
        else:
            i[0] = Gascon.rotr32(i[0],shift2);
            i[1] = Gascon.rotr32(i[1],shift2);
        val = (i[1]<<32) | i[0]
        return val


if __name__ == "__main__":

    def sbox_nw7_equations1(i):
        x0 = (i >> 0) & 1
        x1 = (i >> 1) & 1
        x2 = (i >> 2) & 1
        x3 = (i >> 3) & 1
        x4 = (i >> 4) & 1
        x5 = (i >> 5) & 1
        x6 = (i >> 6) & 1

        a0 = x0 ^ x6
        a1 = x1
        a2 = x2 ^ x1
        a3 = x3
        a4 = x4 ^ x3
        a5 = x5
        a6 = x6 ^ x5

        b0 = (1 ^ a0) & a1
        b1 = (1 ^ a1) & a2
        b2 = (1 ^ a2) & a3
        b3 = (1 ^ a3) & a4
        b4 = (1 ^ a4) & a5
        b5 = (1 ^ a5) & a6
        b6 = (1 ^ a6) & a0

        c0 = a0 ^ b1
        c1 = a1 ^ b2
        c2 = a2 ^ b3
        c3 = a3 ^ b4
        c4 = a4 ^ b5
        c5 = a5 ^ b6
        c6 = a6 ^ b0

        d0 = c0 ^ c6
        d1 = c1 ^ c0
        d2 = c2
        d3 = c3 ^ c2
        d4 = c4
        d5 = c5 ^ c4
        d6 = c6

        y0 = d0
        y1 = d1
        y2 = d2
        y3 = d3 ^ 1
        y4 = d4
        y5 = d5
        y6 = d6

        o = y0
        o |= y1<<1
        o |= y2<<2
        o |= y3<<3
        o |= y4<<4
        o |= y5<<5
        o |= y6<<6
        return o

    def sbox_nw7_equations2(i):
        x0 = (i >> 0) & 1
        x1 = (i >> 1) & 1
        x2 = (i >> 2) & 1
        x3 = (i >> 3) & 1
        x4 = (i >> 4) & 1
        x5 = (i >> 5) & 1
        x6 = (i >> 6) & 1

        b0 = ((1 ^ x0 ^ x6) & x1)
        b1 = ((1 ^ x1) & (x2 ^ x1))
        b2 = ((1 ^ x2 ^ x1) & x3)
        b3 = ((1 ^ x3) & (x4 ^ x3))
        b4 = ((1 ^ x4 ^ x3) & x5)
        b5 = ((1 ^ x5) & (x6 ^ x5))
        b6 = ((1 ^ x6 ^ x5) & (x0 ^ x6))

        c0 = x0 ^ x6 ^ (b1)
        c1 = x1 ^ (b2)
        c2 = x2 ^ x1 ^ (b3)
        c3 = x3 ^ (b4)
        c4 = x4 ^ x3 ^ (b5)
        c5 = x5 ^ (b6)
        c6 = x6 ^ x5 ^ (b0)

        d0 = c0 ^ c6
        d1 = c1 ^ c0
        d2 = c2
        d3 = c3 ^ c2
        d4 = c4
        d5 = c5 ^ c4
        d6 = c6

        y0 = d0
        y1 = d1
        y2 = d2
        y3 = d3 ^ 1
        y4 = d4
        y5 = d5
        y6 = d6

        o = y0
        o |= y1<<1
        o |= y2<<2
        o |= y3<<3
        o |= y4<<4
        o |= y5<<5
        o |= y6<<6
        return o

    def sbox_nw7_equations3(i):
        x0 = (i >> 0) & 1
        x1 = (i >> 1) & 1
        x2 = (i >> 2) & 1
        x3 = (i >> 3) & 1
        x4 = (i >> 4) & 1
        x5 = (i >> 5) & 1
        x6 = (i >> 6) & 1

        c0 = x0 ^ x6 ^ ((1 ^ x1) & (x2 ^ x1))
        c1 = x1 ^ ((1 ^ x2 ^ x1) & x3)
        c2 = x2 ^ x1 ^ ((1 ^ x3) & (x4 ^ x3))
        c3 = x3 ^ ((1 ^ x4 ^ x3) & x5)
        c4 = x4 ^ x3 ^ ((1 ^ x5) & (x6 ^ x5))
        c5 = x5 ^ ((1 ^ x6 ^ x5) & (x0 ^ x6))
        c6 = x6 ^ x5 ^ ((1 ^ x0 ^ x6) & x1)

        d0 = c0 ^ c6
        d1 = c1 ^ c0
        d2 = c2
        d3 = c3 ^ c2
        d4 = c4
        d5 = c5 ^ c4
        d6 = c6

        y0 = d0
        y1 = d1
        y2 = d2
        y3 = d3 ^ 1
        y4 = d4
        y5 = d5
        y6 = d6

        o = y0
        o |= y1<<1
        o |= y2<<2
        o |= y3<<3
        o |= y4<<4
        o |= y5<<5
        o |= y6<<6
        return o

    def sbox_nw7_equations4(i):
        x0 = (i >> 0) & 1
        x1 = (i >> 1) & 1
        x2 = (i >> 2) & 1
        x3 = (i >> 3) & 1
        x4 = (i >> 4) & 1
        x5 = (i >> 5) & 1
        x6 = (i >> 6) & 1

        d0 = x0 ^ x6 ^ ((1 ^ x1) & (x2 ^ x1)) ^ x6 ^ x5 ^ ((1 ^ x0 ^ x6) & x1)
        d1 = x1 ^ ((1 ^ x2 ^ x1) & x3) ^ x0 ^ x6 ^ ((1 ^ x1) & (x2 ^ x1))
        d2 = x2 ^ x1 ^ ((1 ^ x3) & (x4 ^ x3))
        d3 = x3 ^ ((1 ^ x4 ^ x3) & x5) ^ x2 ^ x1 ^ ((1 ^ x3) & (x4 ^ x3))
        d4 = x4 ^ x3 ^ ((1 ^ x5) & (x6 ^ x5))
        d5 = x5 ^ ((1 ^ x6 ^ x5) & (x0 ^ x6)) ^ x4 ^ x3 ^ ((1 ^ x5) & (x6 ^ x5))
        d6 = x6 ^ x5 ^ ((1 ^ x0 ^ x6) & x1)

        y0 = d0
        y1 = d1
        y2 = d2
        y3 = d3 ^ 1
        y4 = d4
        y5 = d5
        y6 = d6

        o = y0
        o |= y1<<1
        o |= y2<<2
        o |= y3<<3
        o |= y4<<4
        o |= y5<<5
        o |= y6<<6
        return o

    def sbox_nw7_equations5(i):
        x0 = (i >> 0) & 1
        x1 = (i >> 1) & 1
        x2 = (i >> 2) & 1
        x3 = (i >> 3) & 1
        x4 = (i >> 4) & 1
        x5 = (i >> 5) & 1
        x6 = (i >> 6) & 1

        y0 = x0 ^ x6 ^ ((1 ^ x1) & (x2 ^ x1)) ^ x6 ^ x5 ^ ((1 ^ x0 ^ x6) & x1)
        y1 = x1 ^ ((1 ^ x2 ^ x1) & x3) ^ x0 ^ x6 ^ ((1 ^ x1) & (x2 ^ x1))
        y2 = x2 ^ x1 ^ ((1 ^ x3) & (x4 ^ x3))
        y3 = x3 ^ ((1 ^ x4 ^ x3) & x5) ^ x2 ^ x1 ^ ((1 ^ x3) & (x4 ^ x3)) ^ 1
        y4 = x4 ^ x3 ^ ((1 ^ x5) & (x6 ^ x5))
        y5 = x5 ^ ((1 ^ x6 ^ x5) & (x0 ^ x6)) ^ x4 ^ x3 ^ ((1 ^ x5) & (x6 ^ x5))
        y6 = x6 ^ x5 ^ ((1 ^ x0 ^ x6) & x1)

        o = y0
        o |= y1<<1
        o |= y2<<2
        o |= y3<<3
        o |= y4<<4
        o |= y5<<5
        o |= y6<<6
        return o

    def sbox_nw7_equations(i):
        x0 = (i >> 0) & 1
        x1 = (i >> 1) & 1
        x2 = (i >> 2) & 1
        x3 = (i >> 3) & 1
        x4 = (i >> 4) & 1
        x5 = (i >> 5) & 1
        x6 = (i >> 6) & 1

        y0 = (x6 & x1) ^ x5 ^ (x2 & x1) ^ x2 ^ (x1 & x0) ^ x1 ^ x0
        y1 = x6 ^ (x3 & x2) ^ (x3 & x1) ^ x3 ^ (x2 & x1) ^ x2 ^ x1 ^  x0
        y2 = (x4 & x3) ^ x4 ^ x2 ^ x1
        y3 = (x5 & x4) ^ (x5 & x3) ^ x5 ^ (x4 & x3) ^ x4 ^ x3 ^ x2 ^ x1 ^ 1
        y4 = (x6 & x5) ^ x6 ^ x4 ^ x3
        y5 = (x6 & x0) ^ x6 ^ (x5 & x0) ^ x5  ^ x4 ^ x3 ^ x0
        y6 = (x6 & x1) ^ x6 ^ x5 ^ (x1 & x0) ^ x1

        o = y0
        o |= y1<<1
        o |= y2<<2
        o |= y3<<3
        o |= y4<<4
        o |= y5<<5
        o |= y6<<6
        return o

    def test_nw7_equations():
        nw=7
        for i in range(0,1<<nw):
            r=Gascon.sbox(i,nw)
            o=sbox_nw7_equations(i)
            if(r!=o):
                print("i=%x, r=%x, o=%x"%(i,r,o))
                exit()

    def sbox_nw5_equations(i):
        x0 = (i >> 0) & 1
        x1 = (i >> 1) & 1
        x2 = (i >> 2) & 1
        x3 = (i >> 3) & 1
        x4 = (i >> 4) & 1

        y0 = (x4 & x1) ^ x3 ^ (x2 & x1) ^ x2 ^(x1 & x0) ^ x1 ^ x0
        y1 = x4 ^ (x3 & x2) ^ (x3 & x1) ^ x3 ^ (x2 & x1) ^ x2 ^ x1 ^ x0
        y2 = (x4 & x3) ^ x4 ^ x2 ^ x1 ^  1
        y3 = (x4 & x0) ^ x4 ^ (x3 & x0) ^ x3 ^ x2 ^ x1 ^ x0
        y4 = (x4 & x1) ^ x4 ^ x3 ^ (x1 & x0) ^ x1

        o = y0
        o |= y1<<1
        o |= y2<<2
        o |= y3<<3
        o |= y4<<4
        return o

    def test_bijectivity(nw):
        used = {}
        for i in range(0,1<<nw):
            r=Gascon.sbox(i,nw)
            #print("%4x --> %4x"%(i,r))
            assert(r not in used)
            used[r]=1
        print("nw=%d is bijective"%nw)

    def gen_latex_ref_table(nw):
        for i in range(0,1<<nw):
            r=Gascon.sbox(i,nw)
            if(15==(i%16)):
                print("%02x \\\\"%(r))
            else:
                print("%02x & "%(r),end="")
        print()

    def get_sbox_size(sbox):
        s0 = sbox(0)
        nw=None
        for i in range(1,1<<16):
            if s0==sbox(i):
                nw=i.bit_length()-1
                break
        assert(nw is not None)
        #print(nw)
        return nw

    def gen_raw_table(sbox):
        nw = get_sbox_size(sbox)
        for i in range(0,1<<nw):
            r=sbox(i)
            print("%02x\t"%(r),end="")
        print()

    def gen_bin_table(sbox):
        nw = get_sbox_size(sbox)
        for i in range(0,1<<nw):
            r=sbox(i)
            print("{0:0{width}b} --> {1:0{width}b}".format(i,r,width=nw))
        print()

    def test_cycle(nw):
        s=0
        cnt=0
        while(True):
            n = Gascon.sbox(s,nw)
            s = n
            cnt+=1
            if n==0:
                break
        print("nw= %d: cycle length from 0 = %d"%(nw,cnt))
        return cnt

    def reverse(x,width):
        out = 0
        for i in range(0,width):
            out |= ((x >> i) & 1)<<(width-1-i)
        return out

    def print_reversed_indexes(width):
        [print("0x%02x, "%(reverse(x,width)),end="") for x in range(0,1<<width)]

    def table_to_str(t):
        return '\n'.join([' '.join([str(cell) for cell in row]) for row in t])

    def sub_zeros(str):
        str = str.replace('0', '.')
        return str

    def int_to_color(i):
        if i:
            i = i*4+4
        return "blue!%s"%str(i)

    def table_to_latex_colored(t):
        return ' \\\\\n'.join([' & '.join([" \cellcolor{%s} "%(int_to_color(cell)) for cell in row]) for row in t])

    def table_to_latex(t):
        return ' \\\\\n'.join([' & '.join(["%s"%(cell) for cell in row]) for row in t])+" \\\\\n"

    def table_to_svg_colored(t):
        numrows = len(t)    # 3 rows in your example
        numcols = len(t[0])
        svg = '<svg width="%d" height="%d">\n'%(numcols*10,numrows*10)
  #<rect width="300" height="100" style="fill:rgb(0,0,255);stroke-width:3;stroke:rgb(0,0,0)" />
        for col in range(0,numcols):
            for row in range(0,numrows):
                b = 255
                r = 255
                if t[row][col]:
                    r = max(0,255-(t[row][col] * 4+4))
                g = r
                svg += '<rect width="10" height="10" x="%d" y="%d" style="fill:rgb(%d,%d,%d)" />\n'%(col*10,row*10,r,g,b)
        svg += "</svg>"
        return svg

    def test_nw5_equations():
        nw=5
        for i in range(0,1<<nw):
            r=Gascon.sbox(i,nw)
            o=sbox_nw5_equations(i)
            if(r!=o):
                print("i=%x, r=%x, o=%x"%(i,r,o))
                exit()

    def bitcount(n):
        count = 0
        while n > 0:
            if (n & 1 == 1): count += 1
            n >>= 1
        return count

    def diff_branch_number(sbox):
        n = get_sbox_size(sbox)
        out = 1<<n
        for x0 in range(0,(1<<n) - 1):
            for x1 in range(x0+1,(1<<n)):
                out = min(out,bitcount(x0^x1)+bitcount(sbox(x0)^sbox(x1)))
        return out

    def lin_branch_number(sbox):
        n = get_sbox_size(sbox)
        out = 1<<n
        c = correlation_coef_table(sbox)
        for a in range(0,(1<<n)):
            for b in range(0,(1<<n)):
                cab = c[a][b]
                if cab:
                    out = min(out,bitcount(a)+bitcount(b))
        return out

    def delta_table(sbox,width):
        dim = (1<<width)
        delta = [ [0] * dim for _ in range(dim)]
        bmax=0
        for x in range(0,dim):
            sx = sbox(x)
            for a in range(0,dim):
                sxa = sbox(x^a)
                b = sx^sxa
                bmax = max(bmax,b)
                delta[a][b] += 1
        #print(bmax)
        return delta

    def correlation_coef_table(sbox):
        n = get_sbox_size(sbox)
        dim = (1<<n)
        out = [ [0] * dim for _ in range(dim)]
        for a in range(0,1<<n):
            for b in range(0,1<<n):
                s=0
                for x in range(0,1<<n):
                    ax = bitcount(a & x) & 1
                    bsx = bitcount(b & sbox(x)) & 1
                    e = ax^bsx
                    if 0==e:
                        s+=1
                    else:
                        s-=1
                out[a][b] = s//2
        return out

    def sbox9(i):
        return Gascon.sbox(i,9)

    def proba_key_second_iteration(width):
        import math
        sel_width = 4
        all_diff = 1
        for i in range(0,sel_width):
            all_diff *= 2**width-i
        key_space = 2 ** (4*width)
        colision = key_space - all_diff
        inv_proba = math.ceil(key_space/colision)
        inv_proba_bits = inv_proba.bit_length()
        print("width=%d, colision: %d (%d bits), all_diff keys: %d (%d bits), proba=2^-%d"%(width,colision, colision.bit_length(), all_diff,all_diff.bit_length(),inv_proba_bits))


    proba_key_second_iteration(2)
    proba_key_second_iteration(4)
    proba_key_second_iteration(8)
    proba_key_second_iteration(16)
    proba_key_second_iteration(32)
    #for nw in range(5,9,2):
    #    gen_latex_ref_table(nw)
    #    test_bijectivity(nw)
    #test_nw7_equations()
    test_nw5_equations()
    nw=9
    #gen_latex_ref_table(nw)
    #print(table_to_svg_colored(delta_table(sbox_nw5_equations,5)))
    #gen_raw_table(sbox9)
    gen_bin_table(sbox_nw5_equations)
    #print(diff_branch_number(sbox_nw7_equations))
    #print(sub_zeros(table_to_latex(correlation_coef_table(sbox_nw5_equations))))
    #print(table_to_svg_colored(correlation_coef_table(sbox9)))
    #print(table_to_svg_colored(delta_table(sbox9,9)))
    #print(diff_branch_number(sbox_nw7_equations))
        #test_cycle(nw)




import binascii
import copy
import sys
import math

if __name__ == "__main__":
    from base import DrySponge
    from gascon import Gascon
else:
    from drysponge.base import DrySponge
    from drysponge.gascon import Gascon

class DryGascon(object):
    def __init__(self,key_min_width,nonce_width,rate_width,capacity_width,x_width,init_rounds,rounds,accumulate_factor=2,version=1):
        assert(0==(capacity_width%64))
        self.nw = capacity_width // 64
        assert(1==(self.nw % 2))
        self.InitRounds = init_rounds
        self.Rounds = rounds
        #self.FinalRounds = final_rounds
        self.MinKeyWidth = key_min_width
        self.NonceWidth = nonce_width
        self.RateWidth = rate_width
        self.CapacityWidth = capacity_width
        assert(0==(x_width%32))
        self.xwords = x_width // 32
        self.MprInputWidth = ((self.xwords-1).bit_length())*(self.CapacityWidth//64)
        #print(self.RateWidth,self.xwords,self.MprInputWidth)
        self.MprInputMask = (1<<self.MprInputWidth)-1
        self.XIdxWidth = (self.xwords-1).bit_length()
        self.XIdxMask = (1<<self.XIdxWidth)-1
        self.AccumulateFactor = accumulate_factor
        #self.mprounds = (self.RateWidth+self.MprInputWidth-1)// self.MprInputWidth
        #print(self.mprounds,self.XIdxWidth,self.XIdxMask)
        #self.dsfree = (self.mprounds*self.MprInputWidth-4 >= rate_width)
        #assert(self.dsfree)
        self.mprounds = (self.RateWidth+4+self.MprInputWidth-1)// self.MprInputWidth
        self.ds=None
        self.spy=False
        self.spy_all=False
        self.rounds = None
        self.version=version

    @staticmethod
    def DryGascon128(version=1):
        return DryGascon(128,128,128,320,32*4,12-1,8-1,2,version=version)

    @staticmethod
    def DryGascon256(version=1):
        return DryGascon(256,128,128,576,32*4,12,8,4,version=version)

    def instance(self):
        return DrySponge(self)

    def DomainSeparator(self,pad,domain,finalize):
        self.ds = pad+finalize*2+domain*4

    def MixPhaseRound(self,c,x,ib,biti,ds=None):
        if self.spy:
            DrySponge.print_bytes_as_hex(c)
        ii =   DrySponge.bytes_to_int(ib)
        if ds is not None:
            #print("ds=%x"%ds)
            #print(ib)
            ii |= ds << (len(ib)*8)
            #print("%x"%ii)
        r=(ii>>biti) & self.MprInputMask
        if self.spy:
            print("biti=%d, r=0x%x"%(biti,r))
        for j in range(0,self.CapacityWidth // 64):
            wi = 0
            #wi = r  & 1
            #r = r >> 1
            i = r & self.XIdxMask
            r = r>>self.XIdxWidth
            ci=((j*2)+wi)
            xi=i
            if self.spy:
                print("c[%d]^=x[%d]: c[%d]="%(ci,xi,ci),end="")
                DrySponge.print_bytes_as_hex(c[ci*4:ci*4+4],end="")
                print(" ^ ",end="")
                DrySponge.print_bytes_as_hex(x[xi*4:xi*4+4],end="")
            for k in range(0,4):
                #print(i,j,k,j*4+k,i*4+k)
                c[ci*4+k] ^= x[xi*4+k]
            if self.spy:
                print(" = ",end="")
                DrySponge.print_bytes_as_hex(c[ci*4:ci*4+4])

        return (c,x)

    def MixPhase(self,c,x,i):
        if self.spy:
            print("ds = %x"%self.ds)
        biti=0
        for j in range(0,self.mprounds-1):
            if self.spy:
                print("MixPhaseRound entry %2d "%j,end="")
            (c,x) = self.MixPhaseRound(c,x,i,biti,self.ds)
            if self.spy:
                print("CoreRound entry     %2d "%j,end="")
                DrySponge.print_bytes_as_hex(c)
            biti += self.MprInputWidth
            c = self.CoreRound(c,0)
        if self.spy:
            print("MixPhaseRound entry %2d "%(self.mprounds-1),end="")
        (c,x) = self.MixPhaseRound(c,x,i,biti,self.ds)
        self.ds = None
        return (c,x)

    def CoreRound(self,c,round):
        S = [0]*self.nw
        for i in range(0,self.nw):
            S[i] = DrySponge.bytes_to_int(c[i*8:(i+1)*8])
        gascon = Gascon(self.nw,12)
        gascon.round(S,round,self.spy_all)
        for i in range(0,self.nw):
            c[i*8:(i+1)*8] = DrySponge.int_to_bytes(S[i],64)
        return c

if __name__ == "__main__":
    if 128==int(sys.argv[1]):
        impl = DryGascon.DryGascon128().instance()
    if 256==int(sys.argv[1]):
        impl = DryGascon.DryGascon256().instance()
    def gen_hash_test_vectors():
        print(r"""\newpage
        """)
        print("\subsection{DryGascon%s hash test vectors}"%sys.argv[1])
        impl.Verbose(DrySponge.SPY_FULL)
        print(r"""\begin{lstlisting}[caption={Detailed test vector}]""")
        impl.hash(binascii.unhexlify(b'0001020304050607'))
        impl.Verbose(DrySponge.SPY_F_IO)
        print(r"""\end{lstlisting}

\begin{lstlisting}[caption={Less detailed test vector}]""")
        impl.hash(binascii.unhexlify(b'0001020304050607'))
        impl.Verbose(DrySponge.SPY_ALG_IO)
        print(r"""\end{lstlisting}

\begin{lstlisting}[caption={Test vectors}]""")
        m=binascii.unhexlify(b'')
        for i in range(0,33):
            impl.hash(m)
            m+=bytes([i])
        impl.Verbose(DrySponge.SPY_NONE)
        iterations = 100
        print(r"""\end{lstlisting}

\begin{lstlisting}[caption={Iterative test vector}]""")
        print("Hashing null message %d times"%iterations)
        m=binascii.unhexlify(b'')
        for i in range(0,iterations):
            m=impl.hash(m)
        print("   Digest: ",end="")
        DrySponge.print_bytes_as_hex(m)
        print(r"""\end{lstlisting}""")

    def gen_aead_test_vectors():
        print("\subsection{DryGascon%s AEAD test vectors}"%sys.argv[1])

        impl.Verbose(DrySponge.SPY_F_IO)
        nonce = binascii.unhexlify(b'F0F1F2F3F4F5F6F7F8F9FAFBFCFDFEFF')[0:impl.nonce_size()]
        #key = binascii.unhexlify(b'000102030405060708090A0B0C0D0E0F101112131415161718191A1B1C1D1E1F202122232425262728292A2B2C2D2E2F')[0:impl.key_size()]
        key = bytearray()
        for i in range(0,impl.fullkey_size()):
            key.append(i)
        print(r"""\begin{lstlisting}[caption={Detailed test vector: m=0, a=0, s=0, full key profile}]""")
        impl.encrypt(key,nonce,binascii.unhexlify(b''),binascii.unhexlify(b''))
        print(r"""\end{lstlisting}""")
        key = key[:impl.fastkey_size()]
        print(r"""\begin{lstlisting}[caption={Detailed test vector: m=0, a=0, s=0, fast key profile}]""")
        impl.encrypt(key,nonce,binascii.unhexlify(b''),binascii.unhexlify(b''))
        print(r"""\end{lstlisting}""")
        key = key[:impl.key_size()]
        print(r"""\begin{lstlisting}[caption={Detailed test vector: m=0, a=0, s=0}]""")
        impl.encrypt(key,nonce,binascii.unhexlify(b''),binascii.unhexlify(b''))
        print(r"""\end{lstlisting}

\begin{lstlisting}[caption={Detailed test vector: m=0, a=0, s=1}]""")
        impl.encrypt(key,nonce,binascii.unhexlify(b''),binascii.unhexlify(b''),binascii.unhexlify(b'BB'))
        print(r"""\end{lstlisting}

\begin{lstlisting}[caption={Detailed test vector: m=0, a=1, s=0}]""")
        impl.encrypt(key,nonce,binascii.unhexlify(b''),binascii.unhexlify(b'AA'))
        print(r"""\end{lstlisting}

\begin{lstlisting}[caption={Detailed test vector: m=1, a=0, s=0}]""")
        impl.encrypt(key,nonce,binascii.unhexlify(b'DD'),binascii.unhexlify(b''))
        print(r"""\end{lstlisting}

\begin{lstlisting}[caption={Detailed test vector: m=1, a=1, s=0}]""")
        impl.encrypt(key,nonce,binascii.unhexlify(b'DD'),binascii.unhexlify(b'AA'))
        print(r"""\end{lstlisting}""")

        impl.Verbose(DrySponge.SPY_ALG_IO)
        print(r"""\begin{lstlisting}[caption={Test vectors}]""")
        m=binascii.unhexlify(b'')
        for i in range(0,33):
            impl.encrypt(key,nonce,m,bytes([0])*16)
            m+=bytes([i])
        impl.Verbose(DrySponge.SPY_NONE)
        print(r"""\end{lstlisting}

\begin{lstlisting}[caption={Iterative test vector}]""")
        iterations = 100
        print("Encrypting null message %d times with tag feedback as associated data"%iterations)
        m=binascii.unhexlify(b'')
        for i in range(0,iterations):
            m=impl.encrypt(key,nonce,binascii.unhexlify(b''),m)
        print("   CipherText: ",end="")
        DrySponge.print_bytes_as_hex(m)
        print(r"""\end{lstlisting}""")

    def test_hash_mode():
        impl.Verbose(DrySponge.SPY_ALG_IO)
        impl.hash(binascii.unhexlify(b'00'))
        impl.hash(binascii.unhexlify(b'0001'))
        impl.hash(binascii.unhexlify(b'0000'))
        impl.hash(binascii.unhexlify(b'000000'))
        impl.hash(binascii.unhexlify(b'01'))
        impl.hash(binascii.unhexlify(b'02'))
        impl.hash(binascii.unhexlify(b'03'))

    def aead_enc_dec(key,nonce,message=None,ad=None,sd=None):
        if message is None:
            message = bytearray(binascii.unhexlify(b''))
        c = impl.encrypt(key,nonce,message,ad,sd)
        p = impl.decrypt(key,nonce,c,ad,sd)
        assert(p==message)

    def debug():
        impl.Verbose(DrySponge.SPY_ALG_IO)
        key = binascii.unhexlify(b'00')*impl.key_size()
        nonce = binascii.unhexlify(b'00')*impl.block_size()
        aead_enc_dec(key,nonce)
        aead_enc_dec(key,nonce,binascii.unhexlify(b'00'))
        aead_enc_dec(key,nonce,binascii.unhexlify(b'0001'))
        aead_enc_dec(key,nonce,binascii.unhexlify(b'0000'))
        aead_enc_dec(key,nonce,binascii.unhexlify(b'000000'))
        aead_enc_dec(key,nonce,binascii.unhexlify(b'00000001'))
        aead_enc_dec(key,nonce,binascii.unhexlify(b'00'),binascii.unhexlify(b'01'),binascii.unhexlify(b'02'))

        #test_hash_mode()
        impl.Verbose(DrySponge.SPY_NONE)
        impl.Verbose(DrySponge.SPY_ALG_IO)
        iterations = 16
        print("Generating expected tag for benchmark on %d iterations"%iterations)
        m=binascii.unhexlify(b'00')*impl.block_size()
        for i in range(0,iterations):
            o=impl.encrypt(key,nonce,m,binascii.unhexlify(b''))
            m=copy.deepcopy(o[impl.block_size():])
        print("   CipherText: ",end="")
        DrySponge.print_bytes_as_hex(o[impl.block_size():])

    #debug()
    gen_aead_test_vectors()
    gen_hash_test_vectors()



def hash(impl):
    from drysponge.base import DrySponge
    import sys
    import binascii

    messageBytes = binascii.unhexlify(b'')
    for i in range(1,len(sys.argv)):
        if (i==len(sys.argv)-1) & (len(sys.argv[i])%2):
            impl.Verbose(int(sys.argv[i]))
        else:
            messageBytes += binascii.unhexlify(sys.argv[i])
    out = impl.hash(messageBytes)
    DrySponge.print_bytes_as_hex(out)
    return DrySponge.bytes_as_hex(out)



"""DryGASCON reference implementation"""

__version__ = '1.0.9'
__title__ = 'drysponge'
__description__ = 'DryGASCON reference implementation'
__long_description__ = """
DryGASCON reference implementation"""
__uri__ = 'https://github.com/sebastien-riou/DryGASCON'
__doc__ = __description__ + ' <' + __uri__ + '>'
__author__ = 'Sebastien Riou'
# For all support requests, please open a new issue on GitHub
__email__ = 'matic@nimp.co.uk'
__license__ = 'Apache 2.0'
__copyright__ = ''
